""" Data interface for a tree of key-value pairs.
"""

from __future__ import annotations
from pyqt_ext.tree import AbstractTreeItem


class KeyValueTreeItem(AbstractTreeItem):
    
    def __init__(self, key: str | None, value, parent: KeyValueTreeItem | None = None) -> None:
        self._key: str | None = key
        self._value = value
        AbstractTreeItem.__init__(self, parent=parent)

        # recursively build subtree if value is itself a container with key,value access
        if isinstance(value, dict):
            for key in value:
                KeyValueTreeItem(key, value[key], parent=self)
        elif isinstance(value, list):
            for i in range(len(value)):
                # list keys are not explicitly set, they will default to the list index
                KeyValueTreeItem(None, value[i], parent=self)
    
    def __repr__(self):
        if self.is_container():
            return f'{self.key}'
        return f'{self.key}: {self.value}'
    
    @property
    def key(self) -> str | int:
        # if parent is a list, the key is the index of this item in the list
        if self.parent is not None:
            if self.parent.is_list():
                return self.sibling_index
        # otherwise, return the local key
        return self._key
    
    @key.setter
    def key(self, key: str) -> None:
        # update parent dict
        if self.parent is not None:
            if self.parent.is_dict():
                # only allow string keys for dict
                key = str(key)
                # remove old key from parent dict
                if (self.key is not None) and (self.key in self.parent.value):
                    self.parent.value.pop(self.key)
                # make sure key is unique
                if key in self.parent.value:
                    key = unique_name(key, list(self.parent.value.keys()))
                # add new key to parent dict
                self.parent.value[key] = self.value
            elif self.parent.is_list():
                # cannot edit list index
                return
        # update local key
        self._key = key
    
    @property
    def value(self):
        return self._value
    
    @value.setter
    def value(self, value) -> None:
        # remove any children
        children = self.children.copy()
        for child in children:
            self.remove_child(child)

        # update parent dict or list
        if self.parent is not None:
            if self.parent.is_dict() or self.parent.is_list():
                self.parent.value[self.key] = value
        # update local value
        self._value = value

        # recursively build subtree if value is itself a container with key,value access
        if isinstance(value, dict):
            for key in value:
                KeyValueTreeItem(key, value[key], parent=self)
        elif isinstance(value, list):
            for i in range(len(value)):
                # list keys are not explicitly set, they will default to the list index
                KeyValueTreeItem(None, value[i], parent=self)
    
    @AbstractTreeItem.parent.setter
    def parent(self, parent: KeyValueTreeItem | None) -> None:
        if self.parent is parent:
            return
        if (parent is not None) and (not parent.is_container()):
            raise ValueError('Parent must be a container (dict or list).')
        old_parent: KeyValueTreeItem | None = self.parent

        # update item tree
        AbstractTreeItem.parent.fset(self, parent)

        # remove value from old parent container
        if old_parent is not None:
            if old_parent.is_dict():
                for key, value in old_parent.value.items():
                    if (value is self.value) or (value == self.value):
                        old_parent.value.pop(key)
                        break
            elif old_parent.is_list():
                if self.value in old_parent.value:
                    old_parent.value.remove(self.value)
        
        # insert value into new parent container
        if parent is not None:
            if parent.is_dict():
                if self.value not in parent.value.values():
                    # resetting key will insert value into parent dict
                    self.key = str(self.key)
                    # # ensure unique key
                    # if key in parent.value:
                    #     key = unique_name(key, list(parent.value.keys()))
                    # self._key = key
                    # parent.value[self.key] = self.value
            elif parent.is_list():
                if self.value not in parent.value:
                    parent.value.append(self.value)
        
        self.ensure_unique_key()
    
    @AbstractTreeItem.name.getter
    def name(self) -> str:
        return str(self.key)
    
    def is_dict(self):
        return isinstance(self.value, dict)
    
    def is_list(self):
        return isinstance(self.value, list)
    
    def is_container(self):
        return self.is_dict() or self.is_list()
    
    def ensure_unique_key(self):
        if self.parent is None:
            return
        keys = [sibling.key for sibling in self.parent.children if sibling is not self]
        if self.key in keys:
            self.key = unique_name(self.key, keys)
    
    def insert_child(self, index: int, item: KeyValueTreeItem) -> bool:
        if not AbstractTreeItem.insert_child(self, index, item):
            return False
        if self.is_list():
            pos = self.value.index(item.value)
            if pos != index:
                self.value.insert(index, self.value.pop(pos))
        return True
    
    def get_data(self, column: int):
        if column == 0:
            return self.key
        elif column == 1:
            return self.value
    
    def set_data(self, column: int, value) -> bool:
        if column == 0:
            self.key = value
            return True
        elif column == 1:
            self.value = value
            return True
        return False


def unique_name(name: str, names: list[str]) -> str:
    if name not in names:
        return name
    i: int = 1
    uname = f'{name}_{i}'
    while uname in names:
        i += 1
        uname = f'{name}_{i}'
    return uname


def test_tree():
    data = {
        'a': 1,
        'b': [4, 8, 9, 5, 7, 99],
        'c': {
            'me': 'hi',
            3: 67,
            'd': {
                'e': 3,
                'f': 'ya!',
                'g': 5,
            },
        },
    }
    root = KeyValueTreeItem('/', data)
    print(root)

    root.remove_child(root['/a'])
    print('remove /a')
    print(root)

    d = root['c/d']
    f = root['c/d/f']
    d.remove_child(f)
    print('remove /c/d/f')
    print(root)

    d.parent = root
    print('move /c/d to /d')
    print(root)

    root.insert_child(1, d)
    print('move /d to 2nd child of /')
    print(root)

    b = root['b']
    b.insert_child(2, d)
    print('move /d to 3rd child of /b')
    print(root)

    b.remove_child(b['1'])
    print('remove /b/1')
    print(root)

    c = root['c']
    c.children[1].value = 'bye'
    print('/c/me:hi -> /c/me:bye')
    print(root)
    print(c)

    c.insert_child(0, b['1'])
    print('move /b/1 to first child of /c')
    print(root)


if __name__ == '__main__':
    test_tree()



""" Data interface and widgets for storing/editing the style of a graph.

Style is stored in hashable dict.
"""

from __future__ import annotations
from qtpy.QtCore import *
from qtpy.QtGui import *
from qtpy.QtWidgets import *
from pyqt_ext.utils import ColorType, toQColor, toColorStr
from pyqt_ext.widgets import ColorButton, CollapsibleSection


class GraphStyle(dict):
    """ Hashable style dict for graph data.

    'color': str
    'linestyle': str
    'linewidth': float
    'marker': str
    'markersize': float
    'markeredgestyle': str
    'markeredgewidth': float
    'markeredgecolor': str
    'markerfacecolor': str
    """

    # matched lists of string reprs and Qt.PenStyles
    linestyles = ['none', '-', '--', ':', '-.', '-..']
    penstyles = [Qt.PenStyle.NoPen, Qt.PenStyle.SolidLine, Qt.PenStyle.DashLine, Qt.PenStyle.DotLine, Qt.PenStyle.DashDotLine, Qt.PenStyle.DashDotDotLine]

    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)

        # alternate key names
        self._keymap = {
            'c': 'color',
            'ls': 'linestyle',
            'lw': 'linewidth',
            'symbol': 'marker',
            'm': 'marker',
            'ms': 'markersize',
            'mes': 'markeredgestyle',
            'mew': 'markeredgewidth',
            'mec': 'markeredgecolor',
            'mfc': 'markerfacecolor',
        }

        # default values
        self._defaults = {
            'linestyle': '-',
            'linewidth': 1,
            'marker': 'none',
            'markersize': 10,
            'markeredgestyle': '-',
            'markeredgewidth': 1,
        }
    
    
    def __getitem__(self, key: str):
        key = key.lower()
        if key in self._keymap:
            key = self._keymap[key]
        if key in self:
            return dict.__getitem__(self, key)
        # key not found...
        if key == 'markeredgewidth':
            if 'linewidth' in self:
                return self['linewidth']
        elif key == 'markeredgecolor':
            if 'color' in self:
                return self['color']
        elif key == 'markerfacecolor':
            if 'markeredgecolor' in self:
                return self['markeredgecolor']
            elif 'color' in self:
                return self['color']
        if key in self._defaults:
            return self._defaults[key]
    
    def __setitem__(self, key: str, value):
        key = key.lower()
        if key in self._keymap:
            key = self._keymap[key]
        if key.endswith('color'):
            if value is not None:
                value = toColorStr(value)
        elif key in ['linestyle', 'markeredgestyle']:
            if value is None:
                value = 'none'
            elif isinstance(value, int):
                value = GraphStyle.linestyles[value]
            elif isinstance(value, Qt.PenStyle):
                value = GraphStyle.linestyles[GraphStyle.penstyles.index(value)]
            elif isinstance(value, str):
                if value == '.-':
                    value = '-.'
                elif value == '..-' or value == '.-.':
                    value = '-..'
        elif key == 'linewidth':
            if value is not None:
                value = max(0, value)
        elif key == 'marker':
            if value is None:
                value = 'none'
        elif key == 'markersize':
            if value is not None:
                value = max(0, value)
        elif key == 'markeredgewidth':
            if value is not None:
                value = max(0, value)
        if value is None:
            del self[key]
            return
        dict.__setitem__(self, key, value)
    
    def __delitem__(self, key: str):
        key = key.lower()
        if key in self._keymap:
            key = self._keymap[key]
        if key in self:
            dict.__delitem__(self, key)
    
    def color(self) -> str | None:
        return self['color']
    
    def setColor(self, value: ColorType | None):
        self['color'] = value
    
    def lineStyle(self) -> str:
        return self['linestyle']
    
    def setLineStyle(self, value: str | int | Qt.PenStyle | None):
        self['linestyle'] = value
    
    def lineWidth(self) -> float:
        return self['linewidth']
    
    def setLineWidth(self, value: float):
        self['linewidth'] = value
    
    def marker(self) -> str:
        return self['marker']
    
    def setMarker(self, value: str | None):
        self['marker'] = value
    
    def markerSize(self) -> float:
        return self['markersize']
    
    def setMarkerSize(self, value: float):
        self['markersize'] = value
    
    def markerEdgeStyle(self) -> str:
        return self['markeredgestyle']
    
    def setMarkerEdgeStyle(self, value: str | int | Qt.PenStyle | None):
        self['markeredgestyle'] = value
    
    def markerEdgeWidth(self) -> float:
        return self['markeredgewidth']
    
    def setMarkerEdgeWidth(self, value: float):
        self['markeredgewidth'] = value
    
    def markerEdgeColor(self) -> str:
        return self['markeredgecolor']
    
    def setMarkerEdgeColor(self, value: ColorType | None):
        self['markeredgecolor'] = value
    
    def markerFaceColor(self) -> str:
        return self['markerfacecolor']
    
    def setMarkerFaceColor(self, value: ColorType | None):
        self['markerfacecolor'] = value


class GraphStylePanel(QWidget):

    def __init__(self, styles: list[str] = None, *args, **kwargs):
        QWidget.__init__(self, *args, **kwargs)

        if styles is None:
            styles = ['color', 'linestyle', 'linewidth', 'marker', 'markersize', 'markeredgestyle', 'markeredgewidth', 'markeredgecolor', 'markerfacecolor']
        self.styles = [style.lower() for style in styles]

        vbox = QVBoxLayout(self)
        vbox.setContentsMargins(10, 10, 10, 10)
        vbox.setSpacing(10)

        # line
        if set(['color', 'linestyle', 'linewidth']).intersection(self.styles):
            self.lineSection = CollapsibleSection(title='Line')
            form = QFormLayout()
            form.setContentsMargins(10, 10, 10, 10)
            form.setSpacing(10)
            if 'color' in self.styles:
                self.colorButton = ColorButton()
                form.addRow('Color', self.colorButton)
            if 'linestyle' in self.styles:
                self.lineStyleComboBox = QComboBox()
                self._lineStyles = {
                    'No Line': 'none',
                    'Solid Line': '-',
                    'Dash Line': '--',
                    'Dot Line': ':',
                    'Dash Dot Line': '-.',
                    'Dash Dot Dot Line': '-..',
                }
                self.lineStyleComboBox.addItems(list(self._lineStyles.keys()))
                self.lineStyleComboBox.setCurrentIndex(1)
                form.addRow('Style', self.lineStyleComboBox)
            if 'linewidth' in self.styles:
                self.lineWidthSpinBox = QDoubleSpinBox()
                self.lineWidthSpinBox.setMinimum(0)
                self.lineWidthSpinBox.setValue(1)
                form.addRow('Width', self.lineWidthSpinBox)
            self.lineSection.setContentLayout(form)
            vbox.addWidget(self.lineSection)
        
        # marker
        if set(['marker', 'markersize', 'markeredgestyle', 'markeredgewidth', 'markeredgecolor', 'markerfacecolor']).intersection(self.styles):
            self.markerSection = CollapsibleSection(title='Marker')
            form = QFormLayout()
            form.setContentsMargins(10, 10, 10, 10)
            form.setSpacing(10)
            if 'marker' in self.styles:
                self.markerComboBox = QComboBox()
                # pyqtgraph default markers
                self.markerComboBox.addItems(['None', 'Circle', 'Triangle Down', 'Triangle Up', 'Triangle Right', 'Triangle Left', 'Square', 'Diamond', 'Pentagon', 'Hexagon', 'Star', 'Plus', 'Cross'])
                self.markerComboBox.setCurrentIndex(0)
                form.addRow('Symbol', self.markerComboBox)
            if 'markersize' in self.styles:
                self.markerSizeSpinBox = QDoubleSpinBox()
                self.markerSizeSpinBox.setMinimum(0)
                self.markerSizeSpinBox.setValue(10)
                form.addRow('Size', self.markerSizeSpinBox)
            if 'markeredgestyle' in self.styles:
                self.markerEdgeStyleComboBox = QComboBox()
                self._markerEdgeStyles = {
                    'No Line': 'none',
                    'Solid Line': '-',
                    'Dash Line': '--',
                    'Dot Line': ':',
                    'Dash Dot Line': '-.',
                    'Dash Dot Dot Line': '-..',
                }
                self.markerEdgeStyleComboBox.addItems(list(self._markerEdgeStyles.keys()))
                self.markerEdgeStyleComboBox.setCurrentIndex(1)
                form.addRow('Edge Style', self.markerEdgeStyleComboBox)
            if 'markeredgewidth' in self.styles:
                self.markerEdgeWidthSpinBox = QDoubleSpinBox()
                self.markerEdgeWidthSpinBox.setMinimum(0)
                self.markerEdgeWidthSpinBox.setValue(1)
                form.addRow('Edge Width', self.markerEdgeWidthSpinBox)
            if 'markeredgecolor' in self.styles:
                self.markerEdgeColorButton = ColorButton()
                form.addRow('Edge Color', self.markerEdgeColorButton)
            if 'markerfacecolor' in self.styles:
                self.markerFaceColorButton = ColorButton()
                form.addRow('Face Color', self.markerFaceColorButton)
            self.markerSection.setContentLayout(form)
            vbox.addWidget(self.markerSection)
        
        # expand 1st section
        if vbox.count() > 0:
            vbox.itemAt(0).widget().expand()
        
        vbox.addStretch()
    
    def graphStyle(self) -> GraphStyle:
        graphStyle = GraphStyle()
        for style in self.styles:
            if style == 'color':
                color = self.colorButton.color()
                if color is not None:
                    graphStyle.setColor(color)
            elif style == 'linestyle':
                key = self.lineStyleComboBox.currentText()
                lineStyle = self._lineStyles[key]
                graphStyle.setLineStyle(lineStyle)
            elif style == 'linewidth':
                lineWidth = self.lineWidthSpinBox.value()
                graphStyle.setLineWidth(lineWidth)
            elif style == 'marker':
                marker = self.markerComboBox.currentText().lower()
                graphStyle.setMarker(marker)
            elif style == 'markersize':
                markerSize = self.markerSizeSpinBox.value()
                graphStyle.setMarkerSize(markerSize)
            elif style == 'markeredgestyle':
                key = self.markerEdgeStyleComboBox.currentText()
                markerEdgeStyle = self._markerEdgeStyles[key]
                graphStyle.setMarkerEdgeStyle(markerEdgeStyle)
            elif style == 'markeredgewidth':
                markerEdgeWidth = self.markerEdgeWidthSpinBox.value()
                graphStyle.setMarkerEdgeWidth(markerEdgeWidth)
            elif style == 'markeredgecolor':
                markerEdgeColor = self.markerEdgeColorButton.color()
                if markerEdgeColor is not None:
                    graphStyle.setMarkerEdgeColor(markerEdgeColor)
            elif style == 'markerfacecolor':
                markerFaceColor = self.markerFaceColorButton.color()
                if markerFaceColor is not None:
                    graphStyle.setMarkerFaceColor(markerFaceColor)
        return graphStyle
    
    def setGraphStyle(self, graphStyle: GraphStyle):
        for style in self.styles:
            if style == 'color':
                try:
                    color = graphStyle.color()
                    self.colorButton.setColor(color)
                except Exception:
                    pass
            elif style == 'linestyle':
                try:
                    lineStyle = graphStyle.lineStyle()
                    for key, value in self._lineStyles.items():
                        if value.lower() == lineStyle.lower() or key.lower() == lineStyle.lower():
                            lineStyle = key
                            break
                    self.lineStyleComboBox.setCurrentText(lineStyle)
                except Exception:
                    pass
            elif style == 'linewidth':
                try:
                    lineWidth = graphStyle.lineWidth()
                    self.lineWidthSpinBox.setValue(lineWidth)
                except Exception:
                    pass
            elif style == 'marker':
                try:
                    marker = graphStyle.marker()
                    for i, item in enumerate(self.markerComboBox.items()):
                        if item.text().lower() == marker.lower():
                            self.markerComboBox.setCurrentIndex(i)
                            break
                except Exception:
                    pass
            elif style == 'markersize':
                try:
                    markerSize = graphStyle.markerSize()
                    self.markerSizeSpinBox.setValue(markerSize)
                except Exception:
                    pass
            elif style == 'markeredgestyle':
                try:
                    markerEdgeStyle = graphStyle.markerEdgeStyle()
                    for key, value in self._markerEdgeStyles.items():
                        if value.lower() == markerEdgeStyle.lower() or key.lower() == markerEdgeStyle.lower():
                            markerEdgeStyle = key
                            break
                    self.markerEdgeStyleComboBox.setCurrentText(markerEdgeStyle)
                except Exception:
                    pass
            elif style == 'markeredgewidth':
                try:
                    markerEdgeWidth = graphStyle.setMarkerEdgeWidth()
                    self.markerEdgeWidthSpinBox.setValue(markerEdgeWidth)
                except Exception:
                    pass
            elif style == 'markeredgecolor':
                try:
                    markerEdgeColor = graphStyle.markerEdgeColor()
                    self.markerEdgeColorButton.setColor(markerEdgeColor)
                except Exception:
                    pass
            elif style == 'markerfacecolor':
                try:
                    markerFaceColor = graphStyle.markerFaceColor()
                    self.markerFaceColorButton.setColor(markerFaceColor)
                except Exception:
                    pass


def editGraphStyle(graphStyle: GraphStyle, styles: list[str] = None, parent: QWidget = None, title: str = None) -> GraphStyle | None:
    panel = GraphStylePanel(styles)
    panel.layout().setContentsMargins(0, 0, 0, 0)
    panel.setGraphStyle(graphStyle)

    dlg = QDialog(parent)
    vbox = QVBoxLayout(dlg)
    vbox.addWidget(panel)

    btns = QDialogButtonBox()
    btns.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok)
    btns.accepted.connect(dlg.accept)
    btns.rejected.connect(dlg.reject)
    vbox.addWidget(btns)
    vbox.addStretch()

    if title is not None:
        dlg.setWindowTitle(title)
    dlg.setWindowModality(Qt.ApplicationModal)
    if dlg.exec() == QDialog.Accepted:
        return panel.graphStyle()


def test_live():
    app = QApplication()
    ui = GraphStylePanel()
    ui.show()
    # QTimer.singleShot(1000, lambda: print(editGraphStyle(GraphStyle())))
    app.exec()


if __name__ == '__main__':
    test_live()




from __future__ import annotations
from qtpy.QtCore import *
from qtpy.QtGui import *
from qtpy.QtWidgets import *

class GraphMeasurementPanel(QWidget):

    measureTypeChanged = Signal(str)
    requestPreview = Signal()
    requestMeasure = Signal()

    def __init__(self, parent=None):
        QWidget.__init__(self, parent)

        self._measureTypeCombobox = QComboBox()
        self._measureTypeCombobox.addItems(['Mean', 'Median'])
        self._measureTypeCombobox.insertSeparator(self._measureTypeCombobox.count())
        self._measureTypeCombobox.addItems(['Min', 'Max', 'AbsMax'])
        self._measureTypeCombobox.insertSeparator(self._measureTypeCombobox.count())
        self._measureTypeCombobox.addItems(['Peaks'])
        self._measureTypeCombobox.insertSeparator(self._measureTypeCombobox.count())
        self._measureTypeCombobox.addItems(['Standard Deviation', 'Variance'])
        self._measureTypeCombobox.currentIndexChanged.connect(lambda index: self._onMeasureTypeChanged())

        self._measureWithinSelectedRegionsCheckbox = QCheckBox('Measure within selected regions')
        self._measureWithinSelectedRegionsCheckbox.setChecked(True)
        self._measureWithinSelectedRegionsCheckbox.stateChanged.connect(lambda state: self._preview())

        self._measurePerSelectedRegionCheckbox = QCheckBox('Measure for each selected region')
        self._measurePerSelectedRegionCheckbox.setChecked(True)
        self._measureWithinSelectedRegionsCheckbox.setEnabled(not self._measurePerSelectedRegionCheckbox.isChecked)
        self._measurePerSelectedRegionCheckbox.stateChanged.connect(lambda state: self._measureWithinSelectedRegionsCheckbox.setEnabled(Qt.CheckState(state) == Qt.CheckState.Unchecked))
        self._measurePerSelectedRegionCheckbox.stateChanged.connect(lambda state: self._preview())

        self._resultNameEdit = QLineEdit()

        self._previewCheckbox = QCheckBox('Preview')
        self._previewCheckbox.setChecked(True)
        self._previewCheckbox.stateChanged.connect(lambda state: self._preview())

        self._measureButton = QPushButton('Measure')
        self._measureButton.pressed.connect(self._measure)

        self._peakTypeCombobox = QComboBox()
        self._peakTypeCombobox.addItems(['Min', 'Max'])
        self._peakTypeCombobox.setCurrentText('Max')
        self._peakTypeCombobox.currentIndexChanged.connect(lambda index: self._preview())

        self._peakAvgHalfWidthSpinbox = QSpinBox()
        self._peakAvgHalfWidthSpinbox.setValue(0)
        self._peakAvgHalfWidthSpinbox.valueChanged.connect(lambda value: self._preview())

        self._peakThresholdEdit = QLineEdit('0')
        self._peakThresholdEdit.editingFinished.connect(self._preview)

        self._peakWidthGroup = QGroupBox()
        form = QFormLayout(self._peakWidthGroup)
        form.setContentsMargins(3, 3, 3, 3)
        form.setSpacing(3)
        form.setHorizontalSpacing(5)
        form.addRow('Average +/- samples', self._peakAvgHalfWidthSpinbox)

        self._peakGroup = QGroupBox()
        form = QFormLayout(self._peakGroup)
        form.setContentsMargins(3, 3, 3, 3)
        form.setSpacing(3)
        form.setHorizontalSpacing(5)
        form.addRow('Peak type', self._peakTypeCombobox)
        form.addRow('Peak threshold', self._peakThresholdEdit)

        self._resultNameWidget = QWidget()
        form = QFormLayout(self._resultNameWidget)
        form.setContentsMargins(0, 0, 0, 0)
        form.setSpacing(0)
        form.setHorizontalSpacing(5)
        form.addRow('Result name', self._resultNameEdit)

        vbox = QVBoxLayout(self)
        vbox.setContentsMargins(3, 3, 3, 3)
        vbox.setSpacing(5)
        vbox.addWidget(self._measureTypeCombobox)
        vbox.addWidget(self._peakGroup)
        vbox.addWidget(self._peakWidthGroup)
        vbox.addWidget(self._measureWithinSelectedRegionsCheckbox)
        vbox.addWidget(self._measurePerSelectedRegionCheckbox)
        # vbox.addWidget(self._resultNameWidget)
        vbox.addWidget(self._previewCheckbox)
        vbox.addWidget(self._measureButton)
        vbox.addStretch()
        
        self._onMeasureTypeChanged()
    
    def _onMeasureTypeChanged(self):
        measureType = self._measureTypeCombobox.currentText()
        self._peakGroup.setVisible(measureType == 'Peaks')
        self._peakWidthGroup.setVisible(measureType in ['Min', 'Max', 'AbsMax', 'Peaks'])
        self._resultNameEdit.setPlaceholderText(measureType)
        self.measureTypeChanged.emit(measureType)
        # self._preview()
    
    def _preview(self):
        self.requestPreview.emit()
    
    def _measure(self):
        self.requestMeasure.emit()


def test_live():
    app = QApplication()
    ui = GraphMeasurementPanel()
    ui.setWindowTitle(ui.__class__.__name__)
    ui.show()
    app.exec()


if __name__ == '__main__':
    test_live()


""" Base class for a tree model that uses AbstractTreeItem for its data interface.
"""

from __future__ import annotations
from qtpy.QtCore import *
from qtpy.QtGui import *
from qtpy.QtWidgets import *
from pyqt_ext.tree import AbstractTreeItem


class AbstractTreeModel(QAbstractItemModel):
    """ Base class for a tree model that uses AbstractTreeItem for its data interface.

    This class can work as is, but in general you will derive from it
        and reimplement `columnCount`, `flags`, etc. to suit your data.

    For drag-and-drop, reimplement `supportedDropActions` and `flags` as needed.
    """

    def __init__(self, root: AbstractTreeItem = None, parent: QObject = None):
        QAbstractItemModel.__init__(self, parent)
        self._root: AbstractTreeItem = root
        self._row_labels: list = []
        self._column_labels: list = []
    
    def root(self) -> AbstractTreeItem:
        return self._root
    
    def setRoot(self, root: AbstractTreeItem) -> None:
        self.beginResetModel()
        self._root = root
        self.endResetModel()
    
    def rowLabels(self) -> list:
        return self._row_labels
    
    def setRowLabels(self, labels: list) -> None:
        old_labels = self._row_labels
        n_overlap = min(len(labels), len(old_labels))
        first_change = 0
        while (first_change < n_overlap) and (labels[first_change] == old_labels[first_change]):
            first_change += 1
        last_change = max(len(labels), len(old_labels)) - 1
        while (last_change < n_overlap) and (labels[last_change] == old_labels[last_change]):
            last_change -= 1
        self._row_labels = labels
        if first_change <= last_change: 
            self.headerDataChanged.emit(Qt.Orientation.Vertical, first_change, last_change)
    
    def columnLabels(self) -> list | None:
        return self._column_labels
    
    def setColumnLabels(self, labels: list | None) -> None:
        old_labels = self._column_labels
        n_overlap = min(len(labels), len(old_labels))
        first_change = 0
        while (first_change < n_overlap) and (labels[first_change] == old_labels[first_change]):
            first_change += 1
        last_change = max(len(labels), len(old_labels)) - 1
        while (last_change < n_overlap) and (labels[last_change] == old_labels[last_change]):
            last_change -= 1
        self._column_labels = labels
        if first_change <= last_change: 
            self.headerDataChanged.emit(Qt.Orientation.Horizontal, first_change, last_change)
    
    def maxDepth(self):
        root: AbstractTreeItem = self.root()
        if root is None:
            return 0
        return root.branch_max_depth()
    
    def rowCount(self, parent_index: QModelIndex = QModelIndex()) -> int:
        """ Uses `AbstractTreeItem.children` to get the number of children of parent.
        """
        if parent_index.column() > 0:
            return 0
        if not parent_index.isValid():
            parent_item: AbstractTreeItem = self.root()
        else:
            parent_item: AbstractTreeItem = self.itemFromIndex(parent_index)
        if parent_item is None:
            return 0
        return len(parent_item.children)

    # !!! must reimplement if you need more than one column
    def columnCount(self, parent_index: QModelIndex = QModelIndex()) -> int:
        # Defaults to a single column tree.
        return 1

    def itemFromIndex(self, index: QModelIndex = QModelIndex()) -> AbstractTreeItem | None:
        """ Get the item associated with index.
        """
        if not index.isValid():
            return self.root()
        item: AbstractTreeItem = index.internalPointer()
        return item
    
    def indexFromItem(self, item: AbstractTreeItem) -> QModelIndex:
        """ Get the index associated with item.
        """
        if (item is self.root()) or (item.parent is None):
            return QModelIndex()
        row: int = item.sibling_index
        col: int = 0
        return self.createIndex(row, col, item)

    def parent(self, index: QModelIndex = QModelIndex()) -> QModelIndex:
        """ Uses `AbstractTreeItem.parent` to get the parent index.
        """
        if not index.isValid():
            return QModelIndex()
        item: AbstractTreeItem = self.itemFromIndex(index)
        parent_item: AbstractTreeItem = item.parent
        if parent_item is None or parent_item is self.root():
            return QModelIndex()
        return self.indexFromItem(parent_item)

    def index(self, row: int, column: int, parent_index: QModelIndex = QModelIndex()) -> QModelIndex:
        """ Return an index associated with the appropriate AbstractTreeItem.
        """
        if not self.hasIndex(row, column, parent_index):
            return QModelIndex()
        if parent_index.isValid() and parent_index.column() != 0:
            return QModelIndex()
        try:
            parent_item: AbstractTreeItem = self.itemFromIndex(parent_index)
            item: AbstractTreeItem = parent_item.children[row]
            return self.createIndex(row, column, item)
        except IndexError:
            return QModelIndex()

    # !!! you may want to cusomize this
    def flags(self, index: QModelIndex) -> Qt.ItemFlags:
        """ Default item flags.
        
        Supports drag-and-drop if it is enabled in `supportedDropActions`.
        """
        if not index.isValid():
            # root item
            if self.supportedDropActions() != Qt.DropAction.IgnoreAction:
                # allow drops on the root item (i.e., this allows drops on the viewport away from other items)
                return Qt.ItemFlag.ItemIsDropEnabled
            return Qt.ItemFlag.NoItemFlags
        flags = Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsSelectable | Qt.ItemFlag.ItemIsEditable
        if self.supportedDropActions() != Qt.DropAction.IgnoreAction:
            flags |= Qt.ItemFlag.ItemIsDragEnabled | Qt.ItemFlag.ItemIsDropEnabled
        return flags

    def data(self, index: QModelIndex, role: int):
        """ Get data via `AbstractTreeItem.get_data`.
        
        Probably do not need to touch this if you have appropriately implemented `AbstractTreeItem.get_data`.
        """
        if not index.isValid():
            return
        item: AbstractTreeItem = self.itemFromIndex(index)
        if role == Qt.ItemDataRole.DisplayRole or role == Qt.ItemDataRole.EditRole:
            return item.get_data(index.column())

    def setData(self, index: QModelIndex, value, role: int) -> bool:
        """ Set data via `AbstractTreeItem.set_data`.
        
        Probably do not need to touch this if you have appropriately implemented `AbstractTreeItem.set_data`.
        """
        item: AbstractTreeItem = self.itemFromIndex(index)
        if role == Qt.ItemDataRole.EditRole:
            success: bool = item.set_data(index.column(), value)
            if success:
                self.dataChanged.emit(index, index)
            return success
        return False

    def headerData(self, section: int, orientation: Qt.Orientation, role: int):
        """ Get data from `rowLabels` or `columnLabels`.
        """
        if role == Qt.ItemDataRole.DisplayRole:
            if orientation == Qt.Orientation.Horizontal:
                labels = self.columnLabels()
            elif orientation == Qt.Orientation.Vertical:
                labels = self.rowLabels()
            if section < len(labels):
                label = labels[section]
                if label is not None:
                    return label
            return section

    def setHeaderData(self, section: int, orientation: Qt.Orientation, value, role: int) -> bool:
        """ Set data in `rowLabels` or `columnLabels`.
        """
        if role == Qt.ItemDataRole.EditRole:
            if orientation == Qt.Orientation.Horizontal:
                labels = self.columnLabels()
            elif orientation == Qt.Orientation.Vertical:
                labels = self.rowLabels()
            if section < len(labels):
                labels[section] = value
            else:
                labels += [None] * (section - len(labels)) + [value]
            if orientation == Qt.Orientation.Horizontal:
                self.setColumnLabels(labels)
            elif orientation == Qt.Orientation.Vertical:
                self.setRowLabels(labels)
            self.headerDataChanged.emit(orientation, section, section)
            return True
        return False

    def removeRows(self, row: int, count: int, parent_index: QModelIndex = QModelIndex()) -> bool:
        """ Calls `AbstractTreeItem.remove_child` to remove rows.
        
        Probably do not need to touch this if you have appropriately implemented `AbstractTreeItem.remove_child`.
        """
        if count <= 0:
            return False
        n_rows: int = self.rowCount(parent_index)
        if row < 0:
            # negative indexing
            row += n_rows
        if (row < 0) or (row + count > n_rows):
            return False
        parent_item: AbstractTreeItem = self.itemFromIndex(parent_index)
        self.beginRemoveRows(parent_index, row, row + count - 1)
        for _ in range(count):
            item: AbstractTreeItem = parent_item.children[row]
            parent_item.remove_child(item)
        self.endRemoveRows()
        return True
    
    def removeItem(self, item: AbstractTreeItem) -> bool:
        """ Remove the item from the model.
        
        Calls `removeRows` internally.
        """
        parent_item: AbstractTreeItem = item.parent
        if parent_item is None:
            # cannot remove the root item
            return False
        parent_index: QModelIndex = self.indexFromItem(parent_item)
        item_row: int = item.sibling_index
        return self.removeRows(item_row, 1, parent_index)
    
    # !!! not implemented
    def insertRows(self, row: int, count: int, parent_index: QModelIndex = QModelIndex()) -> bool:
        """ See `insertItems` instead.
        
        In order to implement this, the model would need to know the type of item to insert (i.e. as derived from AbstractTreeItem).
        """
        return False
    
    def insertItems(self, row: int, items: list[AbstractTreeItem], parent_index: QModelIndex = QModelIndex()) -> bool:
        """ Calls `AbstractTreeItem.insert_child` to insert items (i.e., rows).
        
        Probably do not need to touch this if you have appropriately implemented `AbstractTreeItem.insert_child`.
        """
        if not items:
            return False
        n_rows: int = self.rowCount(parent_index)
        if row < 0:
            # negative indexing
            row += n_rows
        if (row < 0) or (row > n_rows):
            return False
        parent_item: AbstractTreeItem = self.itemFromIndex(parent_index)
        count: int = len(items)
        self.beginInsertRows(parent_index, row, row + count - 1)
        for i, item in enumerate(items):
            parent_item.insert_child(row + i, item)
        self.endInsertRows()
        return True
    
    def appendItems(self, items: list[AbstractTreeItem], parent_index: QModelIndex = QModelIndex()) -> bool:
        row = self.rowCount(parent_index)
        return self.insertItems(row, items, parent_index)
    
    def moveRows(self, src_parent_index: QModelIndex, src_row: int, count: int, dst_parent_index: QModelIndex, dst_row: int) -> bool:
        """ !!! For now this is non-optimial and just repeatedly calls `moveRow` one item at a time.
        """
        n_moved: int = 0
        for i in reversed(list(range(src_row, src_row + count))):
            n_moved += self.moveRow(src_parent_index, i, dst_parent_index, dst_row)
        return n_moved > 0
    
    def moveRow(self, src_parent_index: QModelIndex, src_row: int, dst_parent_index: QModelIndex, dst_row: int) -> bool:
        """ Calls `AbstractTreeItem.insert_child` to move an item (e.g., row) within the tree.
        
        Probably do not need to touch this if you have appropriately implemented `AbstractTreeItem.insert_child`.
        """
        n_src_rows: int = self.rowCount(src_parent_index)
        n_dst_rows: int = self.rowCount(dst_parent_index)
        if src_row < 0:
            # negative indexing
            src_row += n_src_rows
        if dst_row < 0:
            # negative indexing
            dst_row += n_dst_rows
        if (src_parent_index == dst_parent_index) and (src_row == dst_row):
            # no change
            return False
        if not (0 <= src_row < n_src_rows):
            return False
        if not (0 <= dst_row <= n_dst_rows):
            return False

        src_parent_item: AbstractTreeItem = self.itemFromIndex(src_parent_index)
        src_item: AbstractTreeItem = src_parent_item.children[src_row]
        dst_parent_item: AbstractTreeItem = self.itemFromIndex(dst_parent_index)
        if dst_parent_item.has_ancestor(src_item):
            # cannot move an item to one of its descendants
            return False
        if src_parent_item is dst_parent_item:
            if (src_row == dst_row) or (src_row + 1 == dst_row):
                # attempt to move to the same position
                return False

        self.beginMoveRows(src_parent_index, src_row, src_row, dst_parent_index, dst_row)
        dst_parent_item.insert_child(dst_row, src_item)
        self.endMoveRows()
        return True
    
    # !!! reimplement for drag-and-drop
    def supportedDropActions(self) -> Qt.DropActions:
        return Qt.DropAction.IgnoreAction


class AbstractDndTreeModel(AbstractTreeModel):
    """ An AbstractTreeModel that supports drag-and-drop.
    """

    def __init__(self, root: AbstractTreeItem = None, parent: QObject = None):
        AbstractTreeModel.__init__(self, root, parent)
    
    def supportedDropActions(self) -> Qt.DropActions:
        return Qt.DropAction.MoveAction | Qt.DropAction.CopyAction


def test_model():
    root = AbstractTreeItem()
    AbstractTreeItem(parent=root)
    root.append_child(AbstractTreeItem(name='child2'))
    root.insert_child(1, AbstractTreeItem(name='child3'))
    root.children[1].append_child(AbstractTreeItem())
    grandchild2 = AbstractTreeItem(name='grandchild2')
    grandchild2.parent = root['child2']
    AbstractTreeItem(name='greatgrandchild', parent=root['/child2/grandchild2'])

    print('\nInitial model...')
    model = AbstractTreeModel(root)
    print(model.root())

    print('\nRemove grandchild2...')
    model.removeItem(grandchild2)
    print(model.root())

    print('\nInsert grandchild2...')
    model.insertItems(0, [grandchild2], model.indexFromItem(root['child2']))
    print(model.root())


if __name__ == '__main__':
    test_model()


""" Tree model for key-value pairs that uses KeyValueTreeItem for its data interface.
"""

from __future__ import annotations
from qtpy.QtCore import *
from qtpy.QtGui import *
from qtpy.QtWidgets import *
from pyqt_ext.tree import AbstractTreeModel, KeyValueTreeItem
import qtawesome as qta


class KeyValueTreeModel(AbstractTreeModel):

    sigKeyChanged = Signal()
    sigValueChanged = Signal()
    
    def __init__(self, root: KeyValueTreeItem = None, parent: QObject = None):
        AbstractTreeModel.__init__(self, root, parent)
        self.setColumnLabels(['Key', 'Value'])
    
    def columnCount(self, parent: QModelIndex = QModelIndex()) -> int:
        return 2

    def flags(self, index: QModelIndex) -> Qt.ItemFlags:
        if not index.isValid():
            if self.supportedDropActions() != Qt.DropAction.IgnoreAction:
                # allow drops on the root item (i.e., this allows drops on the viewport away from other items)
                return Qt.ItemFlag.ItemIsDropEnabled
            return Qt.ItemFlag.NoItemFlags
        item: KeyValueTreeItem = self.itemFromIndex(index)
        # if (index.column() == 1) and item.is_container():
        #     # cannot edit container value, only the values of items inside it
        #     flags = Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsSelectable
        # else:
        flags = Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsSelectable | Qt.ItemFlag.ItemIsEditable
        if self.supportedDropActions() != Qt.DropAction.IgnoreAction:
            flags |= Qt.ItemFlag.ItemIsDragEnabled
            if item.is_container():
                flags |= Qt.ItemFlag.ItemIsDropEnabled
        # data = self.data(index, Qt.ItemDataRole.DisplayRole)
        # if isinstance(data, bool):
        #     flags |= Qt.ItemFlag.ItemIsUserCheckable
        return flags

    def data(self, index: QModelIndex, role: int):
        if not index.isValid():
            return None
        item: KeyValueTreeItem = self.itemFromIndex(index)
        if role == Qt.ItemDataRole.DisplayRole:
            if index.column() == 1 and item.is_container():
                return None
            return item.get_data(index.column())
        elif role == Qt.ItemDataRole.EditRole:
            return item.get_data(index.column())
        elif role == Qt.ItemDataRole.DecorationRole:
            if index.column() == 0:
                if item.is_dict():
                    return qta.icon('ph.folder-thin')
                if item.is_list():
                    return qta.icon('ph.list-numbers-thin')

    def setData(self, index: QModelIndex, value, role: int) -> bool:
        item: KeyValueTreeItem = self.itemFromIndex(index)
        if role == Qt.ItemDataRole.EditRole:
            success: bool = item.set_data(index.column(), value)
            if success:
                if index.column() == 0:
                    self.sigKeyChanged.emit()
                elif index.column() == 1:
                    self.sigValueChanged.emit()
            return success
        return False


class KeyValueDndTreeModel(KeyValueTreeModel):

    def __init__(self, root: KeyValueTreeItem = None, parent: QObject = None):
        KeyValueTreeModel.__init__(self, root, parent)
    
    def supportedDropActions(self) -> Qt.DropActions:
        return Qt.DropAction.MoveAction | Qt.DropAction.CopyAction


""" Tree view for an AbstractTreeModel with context menu and mouse wheel expand/collapse.
"""

from __future__ import annotations
from qtpy.QtCore import *
from qtpy.QtGui import *
from qtpy.QtWidgets import *
from pyqt_ext.tree import AbstractTreeItem, AbstractTreeModel


class TreeView(QTreeView):
    """ Tree view for an AbstractTreeModel with context menu and mouse wheel expand/collapse.
    """

    CONTEXT_MENU_MAX_LABEL_LENGTH = 50

    selectionWasChanged = Signal()

    def __init__(self, parent: QObject = None) -> None:
        QTreeView.__init__(self, parent)
        sizePolicy = self.sizePolicy()
        sizePolicy.setHorizontalPolicy(QSizePolicy.Policy.Expanding)
        sizePolicy.setVerticalPolicy(QSizePolicy.Policy.Expanding)
        self.setSizePolicy(sizePolicy)
        self.setSizeAdjustPolicy(QAbstractScrollArea.AdjustToContents)
        self.setAlternatingRowColors(True)

        # selection
        self.setSelectionMode(QAbstractItemView.SelectionMode.SingleSelection)
        self.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)

        # context menu
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.onCustomContextMenuRequested)
    
    def setModel(self, model: AbstractTreeModel):
        QTreeView.setModel(self, model)

        # drag and drop?
        is_dnd: bool = model is not None and model.supportedDropActions() != Qt.DropAction.IgnoreAction
        self.setDragEnabled(is_dnd)
        self.setAcceptDrops(is_dnd)
        self.viewport().setAcceptDrops(is_dnd)
        self.setDropIndicatorShown(is_dnd)
        if is_dnd:
            self.setDragDropMode(QAbstractItemView.DragDropMode.DragDrop)
            self.setDefaultDropAction(Qt.DropAction.MoveAction)
        else:
            self.setDragDropMode(QAbstractItemView.DragDropMode.NoDragDrop)
    
    def resetModel(self):
        self.storeState()
        self.model().setRoot(self.model().root())
        self.restoreState()
    
    @Slot(QItemSelection, QItemSelection)
    def selectionChanged(self, selected: QItemSelection, deselected: QItemSelection):
        QTreeView.selectionChanged(self, selected, deselected)
        self.selectionWasChanged.emit()

    def selectedItems(self, column: int | None = 0) -> list[AbstractTreeItem]:
        model: AbstractTreeModel = self.model()
        if model is None:
            return []
        indexes: list[QModelIndex] = self.selectionModel().selectedIndexes()
        if column is not None:
            indexes = [index for index in indexes if index.column() == column]
        items: list[AbstractTreeItem] = [model.itemFromIndex(index) for index in indexes]
        return items
    
    @Slot(QPoint)
    def onCustomContextMenuRequested(self, point: QPoint):
        index: QModelIndex = self.indexAt(point)
        menu: QMenu = self.contextMenu(index)
        if menu is not None:
            menu.exec(self.viewport().mapToGlobal(point))
    
    def contextMenu(self, index: QModelIndex = QModelIndex()) -> QMenu:
        model: AbstractTreeModel = self.model()
        if model is None:
            return None
        
        menu: QMenu = QMenu(self)

        items = self.selectedItems()
        if index.isValid():
            item: AbstractTreeItem = model.itemFromIndex(index)
            label = item.path
            if len(label) > self.CONTEXT_MENU_MAX_LABEL_LENGTH:
                label = '...' + label[-(self.CONTEXT_MENU_MAX_LABEL_LENGTH - 3):]
            item_menu = QMenu(label)
            if index.isValid():
                item_menu.addAction('Delete', lambda self=self, item=item, label=label: self.askToRemoveItem(item, label))
            menu.addMenu(item_menu)
            menu.addSeparator()
            if item in items:
                items.remove(item)
        
        menu.addAction('Expand all', self.expandAll)
        menu.addAction('Collapse all', self.collapseAll)
        if model.columnCount() > 1:
            menu.addAction('Resize columns to contents', self.resizeAllColumnsToContents)
        
        if self.selectionMode() in [QAbstractItemView.ContiguousSelection, QAbstractItemView.ExtendedSelection, QAbstractItemView.MultiSelection]:
            menu.addSeparator()
            menu.addAction('Select all', self.selectAll)
            menu.addAction('Clear selection', self.clearSelection)
        
        if len(items) > 0:
            menu.addSeparator()
            menu.addAction('Remove all selected items', self.removeSelectedItems)
        
        return menu
    
    def expandAll(self):
        QTreeView.expandAll(self)
        try:
            model: AbstractTreeModel = self.model()
            self._depth = model.maxDepth() - 1
        except:
            pass
    
    def collapseAll(self):
        QTreeView.collapseAll(self)
        self._depth = 0
    
    def expandToDepth(self, depth: int):
        model: AbstractTreeModel = self.model()
        if model is not None:
            depth = max(0, min(depth, model.maxDepth() - 1))
            if depth == 0:
                self.collapseAll()
                return
        QTreeView.expandToDepth(self, depth - 1)
        self._depth = depth
    
    def resizeAllColumnsToContents(self):
        model: AbstractTreeModel = self.model()
        if model is None:
            return
        for col in range(model.columnCount()):
            self.resizeColumnToContents(col)
    
    def askToRemoveItem(self, item: AbstractTreeItem, label: str = None):
        if label is None:
            label = item.name
        answer = QMessageBox.question(self, 'Remove', f'Remove {label}?', 
            buttons=QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No, 
            defaultButton=QMessageBox.StandardButton.No
        )
        if answer == QMessageBox.StandardButton.Yes:
            model: AbstractTreeModel = self.model()
            model.removeItem(item)
    
    def removeSelectedItems(self, ask_before_removing: bool = True):
        items: list[AbstractTreeItem] = self.selectedItems()
        if not items:
            return
        if ask_before_removing:
            answer = QMessageBox.question(self, 'Remove', f'Remove selected items?', 
                buttons=QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No, 
                defaultButton=QMessageBox.StandardButton.No
            )
            if answer != QMessageBox.StandardButton.Yes:
                return
        model: AbstractTreeModel = self.model()
        for item in reversed(items):
            model.removeItem(item)
    
    def eventFilter(self, obj: QObject, event: QEvent):
        if event.type() == QEvent.Wheel:
            modifiers: Qt.KeyboardModifier = event.modifiers()
            if Qt.KeyboardModifier.ControlModifier in modifiers \
            or Qt.KeyboardModifier.AltModifier in modifiers \
            or Qt.KeyboardModifier.MetaModifier in modifiers:
                self.mouseWheelEvent(event)
                return True
        # elif event.type() == QEvent.FocusIn:
        #     print('FocusIn')
        #     # self.showDropIndicator(True)
        # elif event.type() == QEvent.FocusOut:
        #     print('FocusOut')
        return QTreeView.eventFilter(self, obj, event)
    
    def mouseWheelEvent(self, event: QWheelEvent):
        delta: int = event.angleDelta().y()
        depth = getattr(self, '_depth', 0)
        if delta > 0:
            self.expandToDepth(depth + 1)
        elif delta < 0:
            self.expandToDepth(depth - 1)
    
    # def mousePressEvent(self, event: QMouseEvent):
    #     if self.selectionMode() != QAbstractItemView.SelectionMode.SingleSelection:
    #         if event.button() == Qt.MouseButton.LeftButton:
    #             index: QModelIndex = self.indexAt(event.pos())
    #             if not index.isValid():
    #                 self.selectionModel().clearSelection()
    #     QTreeView.mousePressEvent(self, event)
    
    def dragEnterEvent(self, event: QDragEnterEvent):
        # print('dragEnterEvent   ', event.mimeData().formats(), event.possibleActions())

        model: AbstractTreeModel = self.model()
        if model is None:
            event.ignore()
            return
        
        index: QModelIndex = event.source().currentIndex()
        if index.isValid() and index != QModelIndex():
            # because we have to clear a stupidly persisting drop indicator on each drop
            self.setDropIndicatorShown(True)

            # Not sure if this is needed?
            self.storeState()

            event.accept()
        else:
            event.ignore()
    
    # def dragMoveEvent(self, event: QDragMoveEvent):
    #     # print('dragMoveEvent')

    #     self.setDropIndicatorShown(True)
    #     QTreeView.dragMoveEvent(self, event)
    
    def dragLeaveEvent(self, event: QDragLeaveEvent):
        # print('dragLeaveEvent')
        pass  # this prevents the drop indicator from being hidden
    
    def dropEvent(self, event: QDropEvent):
        # print('dropEvent')
        
        model: AbstractTreeModel = self.model()
        if model is None:
            event.ignore()
            return
        
        src_index: QModelIndex = event.source().currentIndex()
        if (src_index is None) or (not src_index.isValid()) or (src_index == QModelIndex()):
            event.ignore()
            return
        
        dst_index: QModelIndex = self.indexAt(event.pos())
        if (dst_index is None):
            event.ignore()
            return
        
        dst_parent_index: QModelIndex = model.parent(dst_index)
        dst_row = dst_index.row()

        drop_pos = self.dropIndicatorPosition()
        if drop_pos == QAbstractItemView.DropIndicatorPosition.OnViewport:
            dst_parent_index = QModelIndex()
            dst_row = model.rowCount(dst_parent_index)
        elif drop_pos == QAbstractItemView.DropIndicatorPosition.OnItem:
            dst_parent_index = dst_index
            dst_row = model.rowCount(dst_parent_index)
        elif drop_pos == QAbstractItemView.DropIndicatorPosition.AboveItem:
            pass
        elif drop_pos == QAbstractItemView.DropIndicatorPosition.BelowItem:
            dst_row += 1
        
        src_indices: list[QModelIndex] = [index for index in self.selectionModel().selectedIndexes() if index.column() == 0]
        # handle in reverse so rows are not invalidated
        num_moved = 0
        for src_index in reversed(src_indices):
            if (src_index is None) or (not src_index.isValid()) or (src_index == QModelIndex()):
                continue
        
            src_parent_index: QModelIndex = model.parent(src_index)
            src_row = src_index.row()
        
            if event.dropAction() == Qt.DropAction.MoveAction:
                if self.checkMove(src_parent_index, src_row, dst_parent_index, dst_row):
                    try:
                        success: bool = model.moveRow(src_parent_index, src_row, dst_parent_index, dst_row)
                        num_moved += int(success)
                        if success:
                            if src_parent_index == dst_parent_index:
                                if src_row < dst_row:
                                    dst_row -= 1
                    except Exception as err:
                        QMessageBox.warning(self, 'Move Error', f'{err}')

        # We already handled the drop event, so ignore the default implementation.
        event.ignore()

        # Not sure if this is needed?
        self.restoreState()
        
        # Make sure moved rows are selected.
        self.selectionModel().clearSelection()
        selection: QItemSelection = QItemSelection()
        for row in range(dst_row, dst_row + num_moved):
            index = model.index(row, 0, dst_parent_index)
            selection.merge(QItemSelection(index, index), QItemSelectionModel.SelectionFlag.Select | QItemSelectionModel.SelectionFlag.Rows)
        self.selectionModel().select(selection, QItemSelectionModel.SelectionFlag.Select | QItemSelectionModel.SelectionFlag.Rows)

        # clear persisting drop indicator
        self.setDropIndicatorShown(False)
    
    # def dropMimeData(self, index: QModelIndex, data: QMimeData, action: Qt.DropAction) -> bool:
    #     print('dropMimeData')
    #     return False
    
    # def canDropMimeData(self, data: QMimeData, action: Qt.DropAction, row: int, column: int, parent: QModelIndex) -> bool:
    #     print('canDropMimeData')
    #     return True
    
    def checkMove(self, src_parent_index: QModelIndex, src_row: int, dst_parent_index: QModelIndex, dst_row: int) -> bool:
        return True
    
    def storeState(self):
        model: AbstractTreeModel = self.model()
        if model is None:
            return
        if not hasattr(self, '_state'):
            self._state = {}
        selected: list[QModelIndex] = self.selectionModel().selectedIndexes()
        for item in model.root().depth_first():
            if item is model.root():
                continue
            index: QModelIndex = model.indexFromItem(item)
            path = item.path
            self._state[path] = {
                'expanded': self.isExpanded(index),
                'selected': index in selected
            }

    def restoreState(self):
        model: AbstractTreeModel = self.model()
        if model is None:
            return
        if not hasattr(self, '_state'):
            self._state = {}
        self.selectionModel().clearSelection()
        selection: QItemSelection = QItemSelection()
        for item in model.root().depth_first():
            if item is model.root():
                continue
            index: QModelIndex = model.indexFromItem(item)
            path = item.path
            if path in self._state:
                isExpanded = self._state[path].get('expanded', False)
                self.setExpanded(index, isExpanded)
                isSelected = self._state[path].get('selected', False)
                if isSelected:
                    selection.merge(QItemSelection(index, index), QItemSelectionModel.SelectionFlag.Select | QItemSelectionModel.SelectionFlag.Rows)
            # else:
            #     self.setExpanded(index, False)
        if selection.count():
            self.selectionModel().select(selection, QItemSelectionModel.SelectionFlag.Select | QItemSelectionModel.SelectionFlag.Rows)


def test_live():
    from pyqt_ext.tree import AbstractDndTreeModel

    app = QApplication()

    root = AbstractTreeItem()
    root.append_child(AbstractTreeItem())
    root.append_child(AbstractTreeItem())
    root.append_child(AbstractTreeItem())
    root.children[-1].append_child(AbstractTreeItem())
    root.children[-1].append_child(AbstractTreeItem())
    root.children[-1].append_child(AbstractTreeItem())
    root.children[-1].children[0].append_child(AbstractTreeItem())
    root.children[-1].children[0].append_child(AbstractTreeItem())
    
    model = AbstractDndTreeModel(root)
    view = TreeView()
    view.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
    view.setModel(model)
    view.show()
    view.resize(QSize(600, 600))
    view.expandAll()
    view.resizeAllColumnsToContents()

    app.exec()


if __name__ == '__main__':
    test_live()



""" Base class for a tree of items to be used as a data interface with AbstractTreeModel(QAbstractItemModel).
"""

from __future__ import annotations
from typing import Callable
from collections.abc import Iterator


class AbstractTreeItem():
    """ Base class for a tree of items to be used as a data interface with AbstractTreeModel(QAbstractItemModel).
    
    !!! This only implements the tree structure.
        You must define any data variables in a derived class.
    
    What you get out-of-the-box:
        - A host of methods for navigating and manipulating the tree structure:
            - Parent/child linkage.
            - Host of other linkage properties: siblings, first/last child, first/last sibling, etc.
            - Path-based item access based on item names.
            - Depth-first iteration.
            - String representation of the tree.
            - And more...
        - An interface that is designed to work with AbstractTreeModel (i.e., QAbstractItemModel).

    In a derived class:
        1. Add attributes to store/interface with your data.
        2. Minimally reimplement `get_data` and `set_data` methods (used by AbstractTreeModel).
        3. For nice printout, you may want to reimplement `__repr__`.
        4. For special linkage rules you may need to reimplement `parent.setter` and `insert_child`.
    """
    
    def __init__(self, name: str | None = None, parent: AbstractTreeItem | None = None):
        self.name: str | None = name
        self.parent: AbstractTreeItem | None = parent
        self.children: list[AbstractTreeItem] = []
    
    def __repr__(self) -> str:
        """ Return a single line string representation of this item.
        
        See __str__ for a multi-line representation of the tree.
        """
        return f'name={self.name}, parent={self.parent.name if self.parent else None}, children={[child.name for child in self.children]}'
    
    def __str__(self) -> str:
        """ Returns a multi-line string representation of this item's tree branch.

        Each item is described by its name.
        """
        return self._tree_repr(lambda item: repr(item))
    
    def __getitem__(self, path: str) -> AbstractTreeItem:
        """ Return item at path either from the root item (if path starts with /) or otherwise from this item.
        """
        if path.startswith('/'):
            # path from root item (first name in path must be a child of the root item)
            item: AbstractTreeItem = self.root
        else:
            # path from this item (first name in path must be a child of this item)
            item: AbstractTreeItem = self
        path = path.strip('/').split('/')
        for name in path:
            try:
                child_names = [child.name for child in item.children]
                child_index = child_names.index(name)
                item = item.children[child_index]
            except Exception as error:
                print(error)
                return None
        return item
    
    def dumps(self) -> str:
        """ Returns a multi-line string representation of this item's tree branch.

        Each item is described by its repr.
        """
        return self._tree_repr(lambda item: repr(item))
    
    def _tree_repr(self, func: Callable[[AbstractTreeItem], str]) -> str:
        """ Returns a multi-line string representation of this item's tree branch.

        Each item is described by the single line str returned by func(item).
        See __str__ and dumps for examples.
        """
        items: list[AbstractTreeItem] = list(self.depth_first())
        lines: list[str] = [func(item) for item in items]
        for i, item in enumerate(items):
            if item is self:
                continue
            if item is item.parent.last_child:
                lines[i] = '\u2514' + '\u2500'*2 + ' ' + lines[i]
            else:
                lines[i] = '\u251C' + '\u2500'*2 + ' ' + lines[i]
            parent = item.parent
            while parent is not self:
                if i < items.index(parent.parent.last_child):
                    lines[i] = '\u2502' + ' '*3 + lines[i]
                else:
                    lines[i] = ' '*4 + lines[i]
                parent = parent.parent
        return '\n'.join(lines)
    
    @property
    def parent(self) -> AbstractTreeItem | None:
        return getattr(self, '_parent', None)
    
    @parent.setter
    def parent(self, parent: AbstractTreeItem | None) -> None:
        """ This is the primary method for restructuring the tree.

        Derived classes that need to mirror changes in the tree structure to their data should reimplement parent.setter.
        
        Note: Together, parent.setter and insert_child cover all needed tree restructuring.
        See append_child, insert_child, and remove_child.
        """
        if self.parent is parent:
            # nothing to do
            return
        if (parent is not None) and parent.has_ancestor(self):
            raise ValueError('Cannot set parent to a descendant.')
        if self.parent is not None:
            # detach from old parent
            if self in self.parent.children:
                self.parent.children.remove(self)
        if parent is not None:
            # attach to new parent (appends as last child)
            if self not in parent.children:
                parent.children.append(self)
        self._parent = parent
    
    @property
    def name(self) -> str:
        """ Retun this item's name, or a unique identifier if name is not defined.
        """
        name: str | None = getattr(self, '_name', None)
        if name is None:
            return f'{self.__class__.__name__}@{id(self)}'
        return name
    
    @name.setter
    def name(self, name: str) -> None:
        self._name = name
    
    @property
    def path(self) -> str:
        """ Return the /path/to/this/item from the root item.

        Path is constructed from item names separated by /.
        """
        item: AbstractTreeItem = self
        path = []
        while not item.is_root():
            path.insert(0, item.name)
            item = item.parent
        return '/' + '/'.join(path)
    
    @property
    def root(self) -> AbstractTreeItem:
        item: AbstractTreeItem = self
        while item.parent is not None:
            item = item.parent
        return item
    
    @property
    def first_child(self) -> AbstractTreeItem | None:
        if self.children:
            return self.children[0]

    @property
    def last_child(self) -> AbstractTreeItem | None:
        if self.children:
            return self.children[-1]

    # There is some ambiguity as to whether the following sibling properties should be inclusive of self.
    # To ensure clarity, I have omitted these properties for now.

    # @property
    # def siblings(self) -> list[AbstractTreeItem]:
    #     """ Return a list of this item's siblings (inclusive of this item).
    #     """
    #     if self.parent is not None:
    #         return self.parent.children.copy()
    #     return [self]
    
    # @property
    # def first_sibling(self) -> AbstractTreeItem:
    #     """ Note: Can be this item.
    #     """
    #     if self.parent is not None:
    #         return self.parent.first_child
    #     return self

    # @property
    # def last_sibling(self) -> AbstractTreeItem:
    #     """ Note: Can be this item.
    #     """
    #     if self.parent is not None:
    #         return self.parent.last_child
    #     return self

    @property
    def next_sibling(self) -> AbstractTreeItem | None:
        if self.parent is not None:
            siblings: list[AbstractTreeItem] = self.parent.children
            i: int = siblings.index(self)
            if i+1 < len(siblings):
                return siblings[i+1]

    @property
    def prev_sibling(self) -> AbstractTreeItem | None:
        if self.parent is not None:
            siblings: list[AbstractTreeItem] = self.parent.children
            i: int = siblings.index(self)
            if i-1 >= 0:
                return siblings[i-1]

    @property
    def sibling_index(self) -> int:
        if self.parent is None:
            return 0
        return self.parent.children.index(self)
    
    def depth(self, root: AbstractTreeItem = None) -> int:
        depth: int = 0
        item: AbstractTreeItem = self
        while (item.parent is not None) and (item is not root):
            depth += 1
            item = item.parent
        return depth
    
    def branch_max_depth(self) -> int:
        max_depth: int = 0
        for item in self.depth_first():
            depth: int = item.depth(self)
            if depth > max_depth:
                max_depth = depth
        return max_depth
    
    def is_root(self) -> bool:
        return self.parent is None
    
    def is_leaf(self) -> bool:
        return not self.children
    
    def has_ancestor(self, ancestor: AbstractTreeItem) -> bool:
        item: AbstractTreeItem = self
        while item is not None:
            if item is ancestor:
                return True
            item = item.parent
        return False

    def set_parent(self, parent: AbstractTreeItem):
        self.parent = parent
    
    def append_child(self, child: AbstractTreeItem):
        child.parent = self
    
    def insert_child(self, index: int, child: AbstractTreeItem):
        if not (0 <= index <= len(self.children)):
            return False
        child.parent = self
        # move item to index
        pos = self.children.index(child)
        if pos != index:
            if pos < index:
                index -= 1
            if pos != index:
                self.children.insert(index, self.children.pop(pos))
    
    def remove_child(self, child: AbstractTreeItem):
        if child.parent is not self:
            return False
        child.parent = None
    
    # depth-first iteration --------------------------------------------------
    
    def depth_first(self) -> Iterator[AbstractTreeItem]:
        item: AbstractTreeItem = self
        end_item: AbstractTreeItem | None = self._last_depth_first()._next_depth_first()
        while item is not end_item:
            yield item
            item = item._next_depth_first()
    
    def reverse_depth_first(self) -> Iterator[AbstractTreeItem]:
        item: AbstractTreeItem = self._last_depth_first()
        end_item: AbstractTreeItem | None = self._prev_depth_first()
        while item is not end_item:
            yield item
            item = item._prev_depth_first()
    
    def _next_depth_first(self) -> AbstractTreeItem | None:
        if self.children:
            return self.first_child
        next_sibling: AbstractTreeItem = self.next_sibling
        if next_sibling is not None:
            return next_sibling
        item: AbstractTreeItem = self.parent
        while item is not None:
            next_sibling: AbstractTreeItem = item.next_sibling
            if next_sibling is not None:
                return next_sibling
            item = item.parent
        return None

    def _prev_depth_first(self) -> AbstractTreeItem | None:
        prev_sibling: AbstractTreeItem = self.prev_sibling
        if prev_sibling is not None:
            return prev_sibling._last_depth_first()
        if self.parent is not None:
            return self.parent
        return None
    
    def _last_depth_first(self) -> AbstractTreeItem:
        item: AbstractTreeItem = self
        while item.children:
            item = item.last_child
        return item
    
    # leaf iteration --------------------------------------------------
    
    def leaves(self) -> Iterator[AbstractTreeItem]:
        item: AbstractTreeItem = self._first_leaf()
        end_item: AbstractTreeItem | None = self._last_leaf()._next_leaf()
        while item is not end_item:
            yield item
            item = item._next_leaf()
    
    def reverse_leaves(self) -> Iterator[AbstractTreeItem]:
        item: AbstractTreeItem = self._last_leaf()
        end_item: AbstractTreeItem | None = self._first_leaf()._prev_leaf()
        while item is not end_item:
            yield item
            item = item._prev_leaf()
    
    def _next_leaf(self) -> AbstractTreeItem | None:
        try:
            return self._next_depth_first()._first_leaf()
        except Exception:
            return None

    def _prev_leaf(self) -> AbstractTreeItem | None:
        item: AbstractTreeItem | None = self._prev_depth_first()
        while (item is not None) and item.children:
            item = item._prev_depth_first()
        return item
    
    def _first_leaf(self) -> AbstractTreeItem:
        item: AbstractTreeItem = self
        while item.children:
            item = item.first_child
        return item
    
    def _last_leaf(self) -> AbstractTreeItem:
        item: AbstractTreeItem = self
        while item.children:
            item = item.last_child
        return item

    # interface for QAbstractItemModel ----------------------------------------
    
    def get_data(self, column: int):
        # raise NotImplementedError
        if column == 0:
            # for debugging
            return self.name
    
    def set_data(self, column: int, value) -> bool:
        # raise NotImplementedError
        return False


def test_tree():
    root = AbstractTreeItem()
    AbstractTreeItem(parent=root)
    root.append_child(AbstractTreeItem(name='child2'))
    root.insert_child(1, AbstractTreeItem(name='child3'))
    root.children[1].append_child(AbstractTreeItem())
    grandchild2 = AbstractTreeItem(name='grandchild2')
    grandchild2.parent = root['child2']
    AbstractTreeItem(name='greatgrandchild', parent=root['/child2/grandchild2'])
    
    print('\nInitial tree...')
    print(root)

    print('\nDepth first iteration...')
    for item in root.depth_first():
        print(item.name)

    print('\nRemove grandchild2...')
    grandchild2.parent = None
    print(root)

    print('\nInsert grandchild2...')
    grandchild2.parent = root['child2']
    print(root)


if __name__ == '__main__':
    test_tree()
